@@ -4,7 +4,7 @@ package JSON::Schema::Modern::Utilities;
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Internal utilities for JSON::Schema::Modern
 
-our $VERSION = '0.609';
+our $VERSION = '0.610';
 
 use 5.020;
 use strictures 2;
@@ -325,10 +325,10 @@ sub E ($state, $error_string, @args) {
 
   # we store the absolute uri in unresolved form until needed,
   # and perform the rest of the calculations later.
-  my $uri = [ $state->{initial_schema_uri}, $state->{schema_path}, ($state->{keyword}//()), @schema_path_suffix, $state->{effective_base_uri} ];
+  my $uri = [ $state->@{qw(initial_schema_uri schema_path)}, $state->{keyword}//(), @schema_path_suffix, $state->{effective_base_uri} ];
 
   my $keyword_location = $state->{traversed_schema_path}
-    .jsonp($state->{schema_path}, $state->{keyword}, @schema_path_suffix);
+    .jsonp($state->@{qw(schema_path keyword)}, @schema_path_suffix);
 
   require JSON::Schema::Modern::Error;
   push $state->{errors}->@*, JSON::Schema::Modern::Error->new(
@@ -339,8 +339,8 @@ sub E ($state, $error_string, @args) {
     # we calculate absolute_keyword_location when instantiating the Error object for Result
     _uri => $uri,
     error => @args ? sprintf($error_string, @args) : $error_string,
-    $state->{exception} ? ( exception => $state->{exception} ) : (),
-    $state->{recommended_response} ? ( recommended_response => $state->{recommended_response} ) : (),
+    exception => $state->{exception},
+    ($state->%{recommended_response})x!!$state->{recommended_response},
     mode => $state->{traverse} ? 'traverse' : 'evaluate',
   );
 
@@ -365,10 +365,9 @@ sub A ($state, $annotation) {
 
   # we store the absolute uri in unresolved form until needed,
   # and perform the rest of the calculations later.
-  my $uri = [ $state->{initial_schema_uri}, $state->{schema_path}, $state->{keyword}, $state->{effective_base_uri} ];
+  my $uri = [ $state->@{qw(initial_schema_uri schema_path keyword effective_base_uri)} ];
 
-  my $keyword_location = $state->{traversed_schema_path}
-    .jsonp($state->{schema_path}, $state->{keyword});
+  my $keyword_location = $state->{traversed_schema_path}.jsonp($state->@{qw(schema_path keyword)});
 
   push $state->{annotations}->@*, {
     depth => $state->{depth} // 0,
@@ -480,7 +479,7 @@ JSON::Schema::Modern::Utilities - Internal utilities for JSON::Schema::Modern
 
 =head1 VERSION
 
-version 0.609
+version 0.610
 
 =head1 SYNOPSIS
 
